/*
 * linux/drivers/firmware/pmonvars.c
 *
 * PMON environment variable lister
 *
 * Copyright (C) 2009 Arnaud Patard <apatard@mandriva.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/device.h>
#include <linux/init.h>

extern char **mips_prom_env;

struct pmonvars_attr {
	char *content; /* content of the environment variable */
	struct attribute attr;
	struct list_head list;
};

struct pmonvars_obj {
	struct kobject kobj;
};

#define to_pmonvars_attr(obj) container_of(obj,struct pmonvars_attr,attr)
#define to_pmonvars_obj(obj) container_of(obj,struct pmonvars_obj,kobj)

static struct pmonvars_obj *vars_obj;
static LIST_HEAD(pmonvars_list);

static ssize_t pmonvars_attr_show(struct kobject * kobj, struct attribute *attr, char *buf)
{
	struct pmonvars_attr *env = to_pmonvars_attr(attr);
	return sprintf(buf, "%s\n", env->content);
}

static void pmonvars_release(struct kobject * kobj)
{
	struct pmonvars_obj *obj = to_pmonvars_obj(kobj);
	kfree(obj);
}

static struct sysfs_ops pmonvars_attr_ops = {
	        .show = pmonvars_attr_show,
};

static struct attribute * def_attrs[] = {
	        NULL,
};

static struct kobj_type pmonvars_ktype = {
	.release	= pmonvars_release,
	.sysfs_ops	= &pmonvars_attr_ops,
	.default_attrs	= def_attrs,
};

static struct kset *pmon_kset;

static int pmonvars_add_env(struct pmonvars_obj *pmonvars_obj, char *env)
{
	struct pmonvars_attr *attr;
	char *ptr, *name;
	int err = -ENOMEM;

	if (!env || !*env || !strlen(env)) {
		printk(KERN_WARNING "%s: Empty string ?!\n", __func__);
		return -EINVAL;
	}

	attr = kzalloc(sizeof(struct pmonvars_attr), GFP_KERNEL);
	if (!attr)
		return err;

	ptr = strstr(env, "=");
	if (!ptr) {
		err = -EINVAL;
		goto err_alloc_attr;
	}

	attr->content = kzalloc(strlen(ptr), GFP_KERNEL);
	if (!attr->content) {
		goto err_alloc_attr;
	}
	memcpy(attr->content, ptr+1, strlen(ptr));

	name = kzalloc(ptr-env, GFP_KERNEL);
	if (!name) {
		goto err_alloc_content;
	}
	memcpy(name, env, ptr-env);

	attr->attr.name = name;
	attr->attr.mode = 0444;

	err = sysfs_create_file(&pmonvars_obj->kobj, &attr->attr);
	if (err)
		goto err_alloc_name;

	list_add(&attr->list, &pmonvars_list);
	return 0;

err_alloc_name:
	kfree(name);
err_alloc_content:
	kfree(attr->content);
err_alloc_attr:
	kfree(attr);
	return err;
}

static int __init pmonvars_init(void)
{
	int err, index;

	if (!mips_prom_env)
		return -ENODEV;

	vars_obj = kzalloc(sizeof(struct pmonvars_obj), GFP_KERNEL);
	if (!vars_obj)
		return -ENOMEM;

	pmon_kset = kset_create_and_add("pmon", NULL, firmware_kobj);
	if (!pmon_kset) {
		printk(KERN_ERR "pmonvars: Subsystem registration failed.\n");
		err = -ENOMEM;
		goto err_alloc;
	}

	vars_obj->kobj.kset = pmon_kset;
	err = kobject_init_and_add(&vars_obj->kobj, &pmonvars_ktype, NULL, "env");
	if (err)
		goto err_kset;

	kobject_uevent(&vars_obj->kobj, KOBJ_ADD);

	index = 0;
	while (mips_prom_env[index]) {
		err = pmonvars_add_env(vars_obj, mips_prom_env[index]);
		if (err) {
			printk(KERN_WARNING "%s: Error %d while creating env %d\n",
					__func__, err, index);
		}
		index++;
	}

	return 0;

err_kset:
	kset_unregister(pmon_kset);
err_alloc:
	kfree(vars_obj);
	return err;
}

static void __exit pmonvars_cleanup(void)
{
	struct pmonvars_attr *attr, *n;

	kobject_put(&vars_obj->kobj);
	list_for_each_entry_safe(attr, n, &pmonvars_list, list) {
		list_del(&attr->list);
		kfree(attr->attr.name);
		kfree(attr);
	}
	kset_unregister(pmon_kset);
}

module_init(pmonvars_init);
module_exit(pmonvars_cleanup);


MODULE_AUTHOR("Arnaud Patard <apatard@mandriva.com>");
MODULE_DESCRIPTION("pmon variables");
MODULE_LICENSE("GPL");


