/*
 * drivers/power/lm2e_battery.c
 *
 * Battery driver for Lemote Fulong notebooks.
 *
 * Based on drivers/power/olpc_battery.c
 *
 * Copyright (C) 2007 Arnaud Patard <apatard@mandriva.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * TODO:
 *  - ioremap
 *  - Suppress code duplication with ACPI stuff
 *  - Look at the old smartbatt driver
 *  - get rid of i2c_get_adapter... the whole detection stuff is hacky
 *  and more or less broken
 */

#include <linux/module.h>
#include <linux/err.h>
#include <linux/platform_device.h>
#include <linux/power_supply.h>
#include <linux/jiffies.h>
#include <linux/sched.h>
#include <linux/i2c.h>
#include <linux/timer.h>
#include <asm/mips-boards/bonito64.h>
#include <asm/mips-boards/lm2e_via686.h>

static struct platform_device *bat_pdev;
static int adpt_nr;
#define MAX_I2C_ADPT	6

/*******************************************************************/
/* SMBUS stuff                                                     */
/*******************************************************************/

static void battery_smbus_err_handler(struct i2c_adapter *adapter)
{
	union i2c_smbus_data data;
	int result = 0;
	char *err_str;
	int err_number;

	data.word = 0;

	result = i2c_smbus_xfer(adapter,
		       0xb,
		       0, I2C_SMBUS_READ, 0x16, I2C_SMBUS_BLOCK_DATA, &data);

	err_number = (data.word & 0x000f);

	switch (data.word & 0x000f) {
	case 0x0000:
		err_str = "unexpected bus error";
		break;
	case 0x0001:
		err_str = "busy";
		break;
	case 0x0002:
		err_str = "reserved command";
		break;
	case 0x0003:
		err_str = "unsupported command";
		break;
	case 0x0004:
		err_str = "access denied";
		break;
	case 0x0005:
		err_str = "overflow/underflow";
		break;
	case 0x0006:
		err_str = "bad size";
		break;
	case 0x0007:
		err_str = "unknown error";
		break;
	default:
		err_str = "unrecognized error";
	}
	printk(KERN_INFO"%s: ret %i, err %i\n", err_str, result, err_number);
}

#define	MAX_SMBUS_ERR			1

static int
sbs_smbus_read_word(struct i2c_adapter *adapter, int addr, int func,
			 int * word,
			 void (*err_handler) (struct i2c_adapter *adapter))
{
	union i2c_smbus_data data;
	int result = 0;
	int i;

	if (err_handler == NULL) {
		err_handler = battery_smbus_err_handler;
	}

	for (i = 0; i < MAX_SMBUS_ERR; i++) {
		result = i2c_smbus_xfer(adapter, addr, 0,
							I2C_SMBUS_READ, func,
							I2C_SMBUS_WORD_DATA,
							&data);
		if (result) {
			printk(KERN_INFO"try %i: smbus->adapter.algo->smbus_xfer() failed\n", i);
			if (err_handler) {
				err_handler(adapter);
			}
		} else {
			*word = data.word;
			break;
		}
	}

	return result;
}

static inline void bat_en_charge(void)
{
	VIA686_PM_GPOVAL |= 0x20;
}
static inline void bat_dis_charge(void)
{
	VIA686_PM_GPOVAL &= ~0x20;
}

/*******************************************************************/
/* AC stuff                                                        */
/*******************************************************************/

static int lm2e_ac_get_prop(struct power_supply *psy,
			    enum power_supply_property psp,
			    union power_supply_propval *val)
{
	int ret = 0;
	u32 io_status;

	switch (psp) {
	case POWER_SUPPLY_PROP_ONLINE:
		io_status = VIA686_PM_GPIVAL;
		val->intval  = (io_status & 0x400) ? 1 : 0;
		break;
	default:
		ret = -EINVAL;
		break;
	}
	return ret;
}

static enum power_supply_property lm2e_ac_props[] = {
	POWER_SUPPLY_PROP_ONLINE,
};


static struct power_supply lm2e_ac = {
	.name = "lm2e-ac",
	.type = POWER_SUPPLY_TYPE_MAINS,
	.properties = lm2e_ac_props,
	.num_properties = ARRAY_SIZE(lm2e_ac_props),
	.get_property = lm2e_ac_get_prop,
	//.use_for_apm_ac = 1,
};

/*******************************************************************/
/* SBS stuff                                                       */
/*******************************************************************/

static int lm2e_bat_get_property(struct power_supply *psy,
				 enum power_supply_property psp,
				 union power_supply_propval *val)
{
	int ret = -ENODEV, i;
	u32 io_status;
	u32 curr_bat_cap;
	struct i2c_adapter *adapter;


	io_status = VIA686_PM_GPIVAL;

	/*
	 * Check early if the battery is there.
	 * There's no use try it to find it if it isn't there
	 */
	if (io_status & 0x100) {
		bat_dis_charge();
		if (psp == POWER_SUPPLY_PROP_PRESENT) {
			val->intval = 0;
			return 0;
		}
		else
			return -ENODEV;
	}

	/*
	 * Try to detect where is connected the battery
	 */
	if (adpt_nr == -1) {
		for (i = 0; i < MAX_I2C_ADPT; i++) {
			/* check if there's an adapter
			 * and that it supports smbus xfer
			 */
			adapter = i2c_get_adapter(i);
			if (!adapter || !adapter->algo->smbus_xfer)
				continue;

			/* check if addr is busy */
			ret = i2c_check_addr(adapter, 0xb);

			/* check if there's something */
			if (!ret && i2c_smbus_xfer(adapter, 0x0b, 0, 0, 0, I2C_SMBUS_QUICK, NULL) < 0)
				ret = -ENODEV;
			else
				ret = 0;
			i2c_put_adapter(adapter);
			if (!ret)
				break;
		}
		if (ret)
			return ret;
		adpt_nr = i;
	}

	adapter = i2c_get_adapter(adpt_nr);
	/* this check is a little bit paranoid */
	if (!adapter)
		return -ENODEV;

	if (i2c_smbus_xfer(adapter, 0x0b, 0, 0, 0,
			   I2C_SMBUS_QUICK, NULL) < 0) {
		bat_dis_charge();
		i2c_put_adapter(adapter);
		return -ENODEV;
	}


	ret = sbs_smbus_read_word(adapter, 0xb /*i2c slave addr*/,
					0x0d /*sbs register*/,
					&curr_bat_cap,
					&battery_smbus_err_handler);
	if (ret) {
		printk(KERN_INFO"sbs_adapter_read_word() failed\n");
		bat_dis_charge();
		i2c_put_adapter(adapter);
		return -ENODEV;
	}

	switch (psp) {
	case POWER_SUPPLY_PROP_STATUS:
		if (io_status & 0x400) {
			if (curr_bat_cap < 95) {
				val->intval = POWER_SUPPLY_STATUS_CHARGING;
				bat_en_charge();
			} else {
				val->intval = POWER_SUPPLY_STATUS_FULL;
				bat_dis_charge();
			}
		} else {
			bat_dis_charge();
			val->intval = POWER_SUPPLY_STATUS_DISCHARGING;
		}
		break;

        case POWER_SUPPLY_PROP_CAPACITY:
		val->intval = curr_bat_cap;
		break;

	case POWER_SUPPLY_PROP_CAPACITY_LEVEL:
		if (curr_bat_cap > 95)
			val->intval = POWER_SUPPLY_CAPACITY_LEVEL_FULL;
		else if (curr_bat_cap < 20)
			val->intval = POWER_SUPPLY_CAPACITY_LEVEL_LOW;
		else
			val->intval = POWER_SUPPLY_CAPACITY_LEVEL_NORMAL;
		break;
	default:
		ret = -EINVAL;
		break;
	}

	i2c_put_adapter(adapter);
	return ret;
}

static enum power_supply_property lm2e_bat_props[] = {
	POWER_SUPPLY_PROP_STATUS,
	POWER_SUPPLY_PROP_PRESENT,
	POWER_SUPPLY_PROP_CAPACITY,
	POWER_SUPPLY_PROP_CAPACITY_LEVEL,
};

static struct power_supply lm2e_bat = {
	.properties = lm2e_bat_props,
	.num_properties = ARRAY_SIZE(lm2e_bat_props),
	.get_property = lm2e_bat_get_property,
	.use_for_apm = 1,
};

static int __init lm2e_bat_init(void)
{
	int ret = 0;

	adpt_nr = -1;

	bat_pdev = platform_device_register_simple("lm2e-battery", 0, NULL, 0);
	if (IS_ERR(bat_pdev))
		return PTR_ERR(bat_pdev);

	ret = power_supply_register(&bat_pdev->dev, &lm2e_ac);
	if (ret)
		goto ac_failed;

	lm2e_bat.name = bat_pdev->name;

	ret = power_supply_register(&bat_pdev->dev, &lm2e_bat);
	if (ret)
		goto battery_failed;

	goto success;

battery_failed:
	power_supply_unregister(&lm2e_ac);
ac_failed:
	platform_device_unregister(bat_pdev);
success:
	return ret;
}

static void __exit lm2e_bat_exit(void)
{
	power_supply_unregister(&lm2e_bat);
	power_supply_unregister(&lm2e_ac);
	platform_device_unregister(bat_pdev);
}

module_init(lm2e_bat_init);
module_exit(lm2e_bat_exit);

MODULE_AUTHOR("Arnaud Patard <apatard@mandriva.com>");
MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Battery driver for Lemote notebook");
