/*
 * cs42l51.c
 *
 * ASoC Driver for Cirrus Logic CS42L51 codecs
 *
 * Copyright (c) 2010 Arnaud Patard <apatard@mandriva.com>
 *
 * Based on cs4270.c - Copyright (c) Timur Tabi <timur@freescale.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * For now:
 *  - Only I2C is support. Not SPI
 *  - master mode *NOT* supported
 */

#include <linux/module.h>
#include <linux/platform_device.h>
#include <sound/core.h>
#include <sound/soc.h>
#include <sound/tlv.h>
#include <sound/initval.h>
#include <sound/pcm_params.h>
#include <sound/pcm.h>
#include <linux/i2c.h>

#include "cs42l51.h"

enum funct_mode {
	MODE_SLAVE,
	MODE_SLAVE_AUTO,
	MODE_MASTER,
};

struct cs42l51_private {
	unsigned int mclk;
	unsigned int audio_mode;	/* The mode (I2S or left-justified) */
	enum funct_mode func;		/* See funct_mode */
};

static struct snd_soc_device *cs42l51_socdev;

#define CS42L51_FORMATS ( \
		SNDRV_PCM_FMTBIT_S16_LE  | SNDRV_PCM_FMTBIT_S16_BE  | \
		SNDRV_PCM_FMTBIT_S18_3LE | SNDRV_PCM_FMTBIT_S18_3BE | \
		SNDRV_PCM_FMTBIT_S20_3LE | SNDRV_PCM_FMTBIT_S20_3BE | \
		SNDRV_PCM_FMTBIT_S24_LE  | SNDRV_PCM_FMTBIT_S24_BE)

static int cs42l51_fill_cache(struct snd_soc_codec *codec)
{
	u8 *cache = codec->reg_cache;
	struct i2c_client *i2c_client = codec->control_data;
	s32 length;

	length = i2c_smbus_read_i2c_block_data(i2c_client,
			CS42L51_FIRSTREG | 0x80, CS42L51_NUMREGS, cache);
	if (length != CS42L51_NUMREGS) {
		dev_err(&i2c_client->dev,
				"I2C read failure, addr=0x%x (ret=%d vs %d)\n",
				i2c_client->addr, length, CS42L51_NUMREGS);
		return -EIO;
	}

	return 0;
}

static unsigned int cs42l51_read_reg_cache(struct snd_soc_codec *codec,
	unsigned int reg)
{
	u8 *cache = codec->reg_cache;

	if ((reg < CS42L51_FIRSTREG) || (reg > CS42L51_LASTREG))
		return -EIO;

	return cache[reg - CS42L51_FIRSTREG];
}

static int cs42l51_i2c_write(struct snd_soc_codec *codec, unsigned int reg,
	unsigned int value)
{
	u8 *cache = codec->reg_cache;
	struct i2c_client *client = codec->control_data;

	if ((reg < CS42L51_FIRSTREG) || (reg > CS42L51_LASTREG))
		return -EIO;

	/* Only perform an I2C operation if the new value is different */
	if (cache[reg - CS42L51_FIRSTREG] != value) {
		if (i2c_smbus_write_byte_data(client, reg, value)) {
			dev_err(&client->dev, "I2C write failed\n");
			return -EIO;
		}
		cache[reg - CS42L51_FIRSTREG] = value;
	}

	return 0;
}


static int cs42l51_i2c_probe(struct i2c_client *i2c_client,
	const struct i2c_device_id *id)
{
	struct snd_soc_device *socdev = cs42l51_socdev;
	struct snd_soc_codec *codec = socdev->codec;
	int ret = 0;
	int reg;

	if (codec->control_data)
		return -EBUSY;

	/* Note: codec_dai->codec is NULL here */
	codec->reg_cache = kzalloc(CS42L51_NUMREGS, GFP_KERNEL);
	if (!codec->reg_cache) {
		dev_err(&i2c_client->dev,
				"could not allocate register cache\n");
		ret = -ENOMEM;
		goto error;
	}

	/* Verify that we have a CS42L51 */
	ret = i2c_smbus_read_byte_data(i2c_client, CHIP_REV_ID);
	if (ret < 0) {
		dev_err(&i2c_client->dev, "failed to read I2C\n");
		goto error;
	}

	if (ret != MK_CHIP_REV(CHIP_ID, CHIP_REV)) {
		dev_err(&i2c_client->dev, "Invalid chip id\n");
		ret = -ENODEV;
		goto error;
	}

	dev_info(&i2c_client->dev, "found device at I2C address %X\n",
				i2c_client->addr);

	codec->control_data = i2c_client;
	codec->read = cs42l51_read_reg_cache;
	codec->write = cs42l51_i2c_write;
	codec->reg_cache_size = CS42L51_NUMREGS;

	ret = cs42l51_fill_cache(codec);
	if (ret < 0) {
		dev_err(&i2c_client->dev, "failed to fill register cache\n");
		goto error;
	}

	i2c_set_clientdata(i2c_client, codec);

	/*
	 * DAC configuration (right place to do that ?)
	 * - Use signal processor
	 * - auto mute
	 * - vol changes immediate
	 * - no de-emphasize
	 */
	reg = DAC_CTL_DATA_SEL(1) | DAC_CTL_AMUTE | DAC_CTL_DACSZ(0);
	ret = cs42l51_i2c_write(codec, DAC_CTL, reg);
	if (ret < 0)
		goto error;

	/* Unmute PCM-A & PCM-B and set default vol to */
	ret = cs42l51_i2c_write(codec, PCMA_VOL, 0x60);
	if (ret < 0)
		goto error;

	ret = cs42l51_i2c_write(codec, PCMB_VOL, 0x60);
	if (ret < 0)
		goto error;

	/* default for AOUTx */
	ret = cs42l51_i2c_write(codec, AOUTA_VOL, 0x05);
	if (ret < 0)
		goto error;

	ret = cs42l51_i2c_write(codec, AOUTB_VOL, 0x05);
	if (ret < 0)
		goto error;

	/* route microphone */
	ret = cs42l51_i2c_write(codec, ADC_INPUT, 0xF0);
	if (ret < 0)
		goto error;

	return 0;

error:
	kfree(codec->reg_cache);
	codec->reg_cache = NULL;
	codec->reg_cache_size = 0;
	return ret;
}


static const struct i2c_device_id cs42l51_id[] = {
	{"cs42l51", 0},
	{}
};
MODULE_DEVICE_TABLE(i2c, cs42l51_id);

static struct i2c_driver cs42l51_i2c_driver = {
	.driver = {
		.name = "CS42L51 I2C",
		.owner = THIS_MODULE,
	},
	.id_table = cs42l51_id,
	.probe = cs42l51_i2c_probe,
};

static int cs42l51_add_i2c_device(struct platform_device *pdev,
				const struct cs42l51_setup_data *setup)
{
	struct i2c_board_info info;
	struct i2c_adapter *adapter;
	struct i2c_client *client;
	int ret;

	ret = i2c_add_driver(&cs42l51_i2c_driver);
	if (ret != 0) {
		dev_err(&pdev->dev, "can't add i2c driver\n");
		return ret;
	}

	memset(&info, 0, sizeof(struct i2c_board_info));
	info.addr = setup->i2c_address;
	strlcpy(info.type, "cs42l51", I2C_NAME_SIZE);

	adapter = i2c_get_adapter(setup->i2c_bus);
	if (!adapter) {
		dev_err(&pdev->dev, "can't get i2c adapter %d\n",
			setup->i2c_bus);
		goto err_driver;
	}

	client = i2c_new_device(adapter, &info);
	i2c_put_adapter(adapter);
	if (!client) {
		dev_err(&pdev->dev, "can't add i2c device at 0x%x\n",
			(unsigned int)info.addr);
		goto err_driver;
	}

	return 0;

err_driver:
	i2c_del_driver(&cs42l51_i2c_driver);
	return -ENODEV;
}

static int cs42l51_get_chan_mix(struct snd_kcontrol *kcontrol,
			struct snd_ctl_elem_value *ucontrol)
{
	struct snd_soc_codec *codec = snd_kcontrol_chip(kcontrol);
	unsigned long value = snd_soc_read(codec, PCM_MIXER)&3;

	switch (value) {
	default:
	case 0:
		ucontrol->value.integer.value[0] = 0;
		break;
	case 1:
	case 2:
		ucontrol->value.integer.value[0] = 1;
		break;
	case 3:
		ucontrol->value.integer.value[0] = 2;
		break;
	}

	return 0;
}

#define CHAN_MIX_NORMAL	0x00
#define CHAN_MIX_BOTH	0x55
#define CHAN_MIX_SWAP	0xFF

static int cs42l51_set_chan_mix(struct snd_kcontrol *kcontrol,
			struct snd_ctl_elem_value *ucontrol)
{
	struct snd_soc_codec *codec = snd_kcontrol_chip(kcontrol);
	unsigned char val;

	switch (ucontrol->value.integer.value[0]) {
	default:
	case 0:
		val = CHAN_MIX_NORMAL;
		break;
	case 1:
		val = CHAN_MIX_BOTH;
		break;
	case 2:
		val = CHAN_MIX_SWAP;
		break;
	}

	snd_soc_write(codec, PCM_MIXER, val);

	return 1;
}



int snd_soc_info_volsw_2r_sx(struct snd_kcontrol *kcontrol,
			struct snd_ctl_elem_info *uinfo)
{
	struct soc_mixer_control *mc =
		(struct soc_mixer_control *)kcontrol->private_value;
	int max = mc->max;
	int min = mc->min;

	uinfo->type = SNDRV_CTL_ELEM_TYPE_INTEGER;
	uinfo->count = 2;
	uinfo->value.integer.min = 0;
	uinfo->value.integer.max = max-min;

	return 0;
}

int snd_soc_get_volsw_2r_sx(struct snd_kcontrol *kcontrol,
			struct snd_ctl_elem_value *ucontrol)
{
	struct soc_mixer_control *mc =
		(struct soc_mixer_control *)kcontrol->private_value;
	struct snd_soc_codec *codec = snd_kcontrol_chip(kcontrol);
	unsigned int mask = (1<<mc->shift)-1;
	int min = mc->min;
	int val = snd_soc_read(codec, mc->reg) & mask;
	int valr = snd_soc_read(codec, mc->rreg) & mask;

	ucontrol->value.integer.value[0] = ((val & 0xff)-min);
	ucontrol->value.integer.value[1] = ((valr & 0xff)-min);
	return 0;
}

int snd_soc_put_volsw_2r_sx(struct snd_kcontrol *kcontrol,
			struct snd_ctl_elem_value *ucontrol)
{
	struct soc_mixer_control *mc =
		(struct soc_mixer_control *)kcontrol->private_value;
	struct snd_soc_codec *codec = snd_kcontrol_chip(kcontrol);
	unsigned int mask = (1<<mc->shift)-1;
	int min = mc->min;
	int ret;
	unsigned int val, valr, oval, ovalr;

	val = ((ucontrol->value.integer.value[0]+min) & 0xff);
	val &= mask;
	valr = ((ucontrol->value.integer.value[1]+min) & 0xff);
	valr &= mask;

	oval = snd_soc_read(codec, mc->reg) & mask;
	ovalr = snd_soc_read(codec, mc->rreg) & mask;

	ret = 0;
	if (oval != val) {
		ret = snd_soc_write(codec, mc->reg, val);
		if (ret < 0)
			return 0;
		ret = 1;
	}
	if (ovalr != valr) {
		ret = snd_soc_write(codec, mc->rreg, valr);
		if (ret < 0)
			return 0;
		ret = 1;
	}

	return 0;
}

/* this is unfortunate, but I've to create a new kind of mixer control */
#define SOC_DOUBLE_R_SX_TLV(xname, xreg_left, xreg_right, xshift, xmin, xmax, tlv_array) \
{	.iface = SNDRV_CTL_ELEM_IFACE_MIXER, .name = (xname), \
		.access = SNDRV_CTL_ELEM_ACCESS_TLV_READ | \
			  SNDRV_CTL_ELEM_ACCESS_READWRITE, \
		.tlv.p = (tlv_array), \
		.info = snd_soc_info_volsw_2r_sx, \
		.get = snd_soc_get_volsw_2r_sx, \
		.put = snd_soc_put_volsw_2r_sx, \
		.private_value = (unsigned long)&(struct soc_mixer_control) \
			{.reg = xreg_left, \
			 .rreg = xreg_right, .shift = xshift, \
			 .min = xmin, .max = xmax} }

static const DECLARE_TLV_DB_SCALE(adc_pcm_tlv, -5150, 50, 0);
static const DECLARE_TLV_DB_SCALE(tone_tlv, -1050, 150, 0);
/* This is a lie. after -102 db, it stays at -102 */
/* maybe a range would be better */
static const DECLARE_TLV_DB_SCALE(aout_tlv, -11550, 50, 0);

static const char *mic_boost[] = { "+16dB", "+32dB"};

static const char *chan_mix[] = {
	"L R",
	"L+R",
	"R L",
};

static const struct soc_enum cs42l51_mix[] = {
	SOC_ENUM_SINGLE_EXT(ARRAY_SIZE(chan_mix), chan_mix),
	SOC_ENUM_DOUBLE(MIC_CTL, 0, 1, 2, mic_boost),
};

static const struct snd_kcontrol_new cs42l51_snd_controls[] = {
	SOC_DOUBLE_R_SX_TLV("PCM Playback Volume",
			PCMA_VOL, PCMB_VOL, 7, 0xffffff99, 0x18, adc_pcm_tlv),
	SOC_DOUBLE_R("PCM Playback Switch", PCMA_VOL, PCMB_VOL, 7, 1, 1),
	SOC_DOUBLE_R_SX_TLV("Analog Playback Volume",
			AOUTA_VOL, AOUTB_VOL, 8, 0xffffff19, 0x18, aout_tlv),
	SOC_DOUBLE_R_SX_TLV("ADC Mixer Volume",
			ADCA_VOL, ADCB_VOL, 7, 0xffffff99, 0x18, adc_pcm_tlv),
	SOC_DOUBLE_R("ADC Mixer Switch", PCMA_VOL, PCMB_VOL, 7, 1, 1),
	SOC_SINGLE("Playback Deemphasis", DAC_CTL, 3, 1, 0),
	SOC_SINGLE("Auto-Mute Switch", DAC_CTL, 2, 1, 0),
	SOC_SINGLE("Soft Ramp Switch", DAC_CTL, 1, 1, 0),
	SOC_SINGLE("Zero Cross Switch", DAC_CTL, 0, 0, 0),
	SOC_SINGLE("Mic Bias", MIC_POWER_CTL, 1, 1, 1),
	SOC_DOUBLE("Mic Powerdown", MIC_POWER_CTL, 2, 3, 1, 0),
	SOC_ENUM("Mic Boost", cs42l51_mix[1]),
	SOC_SINGLE_TLV("Bass Volume", TONE_CTL, 0, 0xf, 1, tone_tlv),
	SOC_SINGLE_TLV("Treble Volume", TONE_CTL, 4, 0xf, 1, tone_tlv),
	SOC_ENUM_EXT("PCM channel mixer",
			cs42l51_mix[0],
			cs42l51_get_chan_mix, cs42l51_set_chan_mix),
};

static int cs42l51_set_dai_fmt(struct snd_soc_dai *codec_dai,
		unsigned int format)
{
	struct snd_soc_codec *codec = codec_dai->codec;
	struct cs42l51_private *cs42l51 = codec->private_data;
	int ret = 0;

	switch (format & SND_SOC_DAIFMT_FORMAT_MASK) {
	case SND_SOC_DAIFMT_I2S:
	case SND_SOC_DAIFMT_LEFT_J:
	case SND_SOC_DAIFMT_RIGHT_J:
		cs42l51->audio_mode = format & SND_SOC_DAIFMT_FORMAT_MASK;
		break;
	default:
		printk(KERN_ERR "cs4270: invalid DAI format\n");
		ret = -EINVAL;
	}

	switch (SND_SOC_DAIFMT_MASTER_MASK) {
	default:
	case SND_SOC_DAIFMT_CBM_CFM:
		cs42l51->func = MODE_SLAVE_AUTO;
		break;
	case SND_SOC_DAIFMT_CBS_CFS:
		cs42l51->func = MODE_MASTER;
		break;
	}

	return ret;
}

struct cs42l51_ratios {
	unsigned int ratio;
	unsigned char speed_mode;
	unsigned char mclk;
};

static struct cs42l51_ratios slave_ratios[] = {
	{  512, QSM_MODE, 0 }, {  768, QSM_MODE, 0 }, { 1024, QSM_MODE, 0 },
	{ 1536, QSM_MODE, 0 }, { 2048, QSM_MODE, 0 }, { 3072, QSM_MODE, 0 },
	{  256, HSM_MODE, 0 }, {  384, HSM_MODE, 0 }, {  512, HSM_MODE, 0 },
	{  768, HSM_MODE, 0 }, { 1024, HSM_MODE, 0 }, { 1536, HSM_MODE, 0 },
	{  128, SSM_MODE, 0 }, {  192, SSM_MODE, 0 }, {  256, SSM_MODE, 0 },
	{  384, SSM_MODE, 0 }, {  512, SSM_MODE, 0 }, {  768, SSM_MODE, 0 },
	{  128, DSM_MODE, 0 }, {  192, DSM_MODE, 0 }, {  256, DSM_MODE, 0 },
	{  384, DSM_MODE, 0 },
};

static struct cs42l51_ratios slave_auto_ratios[] = {
	{ 1024, QSM_MODE, 0 }, { 1536, QSM_MODE, 0 }, { 2048, QSM_MODE, 1 },
	{ 3072, QSM_MODE, 1 },
	{  512, HSM_MODE, 0 }, {  768, HSM_MODE, 0 }, { 1024, HSM_MODE, 1 },
	{ 1536, HSM_MODE, 1 },
	{  256, SSM_MODE, 0 }, {  384, SSM_MODE, 0 }, {  512, SSM_MODE, 1 },
	{  768, SSM_MODE, 1 },
	{  128, DSM_MODE, 0 }, {  192, DSM_MODE, 0 }, {  256, DSM_MODE, 1 },
	{  384, DSM_MODE, 1 },
};

/* Fill me */
static struct cs42l51_ratios master_ratios[] = { {},
};

static int cs42l51_set_dai_sysclk(struct snd_soc_dai *codec_dai,
		int clk_id, unsigned int freq, int dir)
{
	struct snd_soc_codec *codec = codec_dai->codec;
	struct cs42l51_private *cs42l51 = codec->private_data;
	struct cs42l51_ratios *ratios = NULL;
	int nr_ratios = 0;
	unsigned int rates = 0;
	unsigned int rate_min = -1;
	unsigned int rate_max = 0;
	int i;

	cs42l51->mclk = freq;

	switch (cs42l51->func) {
	case MODE_MASTER:
		ratios = master_ratios;
		nr_ratios = ARRAY_SIZE(master_ratios);
		break;
	case MODE_SLAVE:
		ratios = slave_ratios;
		nr_ratios = ARRAY_SIZE(slave_ratios);
		break;
	case MODE_SLAVE_AUTO:
		ratios = slave_auto_ratios;
		nr_ratios = ARRAY_SIZE(slave_auto_ratios);
		break;
	}

	for (i = 0; i < nr_ratios; i++) {
		unsigned int rate = freq / ratios[i].ratio;
		rates |= snd_pcm_rate_to_rate_bit(rate);
		if (rate < rate_min)
			rate_min = rate;
		if (rate > rate_max)
			rate_max = rate;
	}
	rates &= ~SNDRV_PCM_RATE_KNOT;

	if (!rates) {
		printk(KERN_ERR "cs42l51: could not find a valid sample rate\n");
		return -EINVAL;
	}

	codec_dai->playback.rates = rates;
	codec_dai->playback.rate_min = rate_min;
	codec_dai->playback.rate_max = rate_max;

	codec_dai->capture.rates = rates;
	codec_dai->capture.rate_min = rate_min;
	codec_dai->capture.rate_max = rate_max;

	return 0;
}

static int cs42l51_hw_params(struct snd_pcm_substream *substream,
		struct snd_pcm_hw_params *params,
		struct snd_soc_dai *dai)
{
	struct snd_soc_pcm_runtime *rtd = substream->private_data;
	struct snd_soc_device *socdev = rtd->socdev;
	struct snd_soc_codec *codec = socdev->codec;
	struct cs42l51_private *cs42l51 = codec->private_data;
	int ret;
	unsigned int i;
	unsigned int rate;
	unsigned int ratio;
	struct cs42l51_ratios *ratios = NULL;
	int nr_ratios = 0;
	int intf_ctl, power_ctl;

	switch (cs42l51->func) {
	case MODE_MASTER:
		ratios = master_ratios;
		nr_ratios = ARRAY_SIZE(master_ratios);
		break;
	case MODE_SLAVE:
		ratios = slave_ratios;
		nr_ratios = ARRAY_SIZE(slave_ratios);
		break;
	case MODE_SLAVE_AUTO:
		ratios = slave_auto_ratios;
		nr_ratios = ARRAY_SIZE(slave_auto_ratios);
		break;
	}

	/* Figure out which MCLK/LRCK ratio to use */
	rate = params_rate(params);     /* Sampling rate, in Hz */
	ratio = cs42l51->mclk / rate;    /* MCLK/LRCK ratio */
	for (i = 0; i < nr_ratios; i++) {
		if (ratios[i].ratio == ratio)
			break;
	}

	if (i == nr_ratios) {
		/* We did not find a matching ratio */
		printk(KERN_ERR "cs42l51: could not find matching ratio\n");
		return -EINVAL;
	}

	intf_ctl = snd_soc_read(codec, INTF_CTL);
	power_ctl = snd_soc_read(codec, MIC_POWER_CTL);

	intf_ctl &= ~(INTF_CTL_MASTER|INTF_CTL_ADC_I2S|INTF_CTL_DAC_FORMAT(7));
	power_ctl &= ~(MIC_POWER_CTL_SPEED(3)|MIC_POWER_CTL_MCLK_DIV2);

	switch (cs42l51->func) {
	case MODE_MASTER:
		intf_ctl |= INTF_CTL_MASTER;
		power_ctl |= MIC_POWER_CTL_SPEED(ratios[i].speed_mode);
		break;
	case MODE_SLAVE:
		power_ctl |= MIC_POWER_CTL_SPEED(ratios[i].speed_mode);
		break;
	case MODE_SLAVE_AUTO:
		power_ctl |= MIC_POWER_CTL_AUTO;
		break;
	}

	switch (cs42l51->audio_mode) {
	case SND_SOC_DAIFMT_I2S:
		intf_ctl |= INTF_CTL_ADC_I2S | INTF_CTL_DAC_FORMAT(DAC_DIF_I2S);
		break;
	case SND_SOC_DAIFMT_LEFT_J:
		intf_ctl |= INTF_CTL_DAC_FORMAT(DAC_DIF_LJ24);
		break;
	case SND_SOC_DAIFMT_RIGHT_J:
		switch (params_format(params)) {
		case SNDRV_PCM_FORMAT_S16_LE:
		case SNDRV_PCM_FORMAT_S16_BE:
			intf_ctl |= INTF_CTL_DAC_FORMAT(DAC_DIF_RJ16);
			break;
		case SNDRV_PCM_FORMAT_S18_3LE:
		case SNDRV_PCM_FORMAT_S18_3BE:
			intf_ctl |= INTF_CTL_DAC_FORMAT(DAC_DIF_RJ18);
			break;
		case SNDRV_PCM_FORMAT_S20_3LE:
		case SNDRV_PCM_FORMAT_S20_3BE:
			intf_ctl |= INTF_CTL_DAC_FORMAT(DAC_DIF_RJ20);
			break;
		case SNDRV_PCM_FORMAT_S24_LE:
		case SNDRV_PCM_FORMAT_S24_BE:
			intf_ctl |= INTF_CTL_DAC_FORMAT(DAC_DIF_RJ24);
			break;
		default:
			printk(KERN_ERR "cs42l51: unknown format\n");
			return -EINVAL;
		}
		break;
	default:
		printk(KERN_ERR "cs42l51: unknown format\n");
		return -EINVAL;
	}

	if (ratios[i].mclk)
		power_ctl |= MIC_POWER_CTL_MCLK_DIV2;

	ret = snd_soc_write(codec, INTF_CTL, intf_ctl);
	if (ret < 0)
		return ret;

	ret = snd_soc_write(codec, MIC_POWER_CTL, power_ctl);
	if (ret < 0)
		return ret;

	return 0;
}

struct snd_soc_dai cs42l51_dai = {
	.name = "CS42L51 HiFi",
	.playback = {
		.stream_name = "Playback",
		.channels_min = 1,
		.channels_max = 2,
		.rates = SNDRV_PCM_RATE_8000_96000,
		.formats = CS42L51_FORMATS,
	},
	.capture = {
		.stream_name = "Capture",
		.channels_min = 1,
		.channels_max = 2,
		.rates = SNDRV_PCM_RATE_8000_96000,
		.formats = CS42L51_FORMATS,
	},
};
EXPORT_SYMBOL_GPL(cs42l51_dai);


static int cs42l51_probe(struct platform_device *pdev)
{
	struct snd_soc_device *socdev = platform_get_drvdata(pdev);
	struct snd_soc_codec *codec;
	struct cs42l51_setup_data *setup;
	int ret = 0;
	int i;

	dev_info(&pdev->dev, "CS42L51 ALSA SoC Codec\n");

	setup = socdev->codec_data;

	/* Allocate enough space for the snd_soc_codec structure
	 * and our private data together.
	 */
	codec = kzalloc(ALIGN(sizeof(struct snd_soc_codec), 4) +
			sizeof(struct cs42l51_private), GFP_KERNEL);
	if (!codec) {
		dev_err(&pdev->dev, "Could not allocate codec structure\n");
		return -ENOMEM;
	}

	mutex_init(&codec->mutex);
	INIT_LIST_HEAD(&codec->dapm_widgets);
	INIT_LIST_HEAD(&codec->dapm_paths);

	codec->name = "CS42L51";
	codec->owner = THIS_MODULE;
	codec->dai = &cs42l51_dai;
	codec->num_dai = 1;
	codec->private_data = (void *) codec +
		ALIGN(sizeof(struct snd_soc_codec), 4);

	socdev->codec = codec;

	/* Register PCMs */
	ret = snd_soc_new_pcms(socdev, SNDRV_DEFAULT_IDX1, SNDRV_DEFAULT_STR1);
	if (ret < 0) {
		dev_err(&pdev->dev, "failed to create PCMs\n");
		goto error_free_codec;
	}

	for (i = 0; i < ARRAY_SIZE(cs42l51_snd_controls); i++) {
		ret = snd_ctl_add(codec->card,
				snd_soc_cnew(&cs42l51_snd_controls[i],
				codec, NULL));
		if (ret < 0) {
			dev_err(&pdev->dev, "failed to add controls\n");
			goto error_free_pcms;
		}
	}

	cs42l51_socdev = socdev;

	ret = cs42l51_add_i2c_device(pdev, setup);
	if (ret) {
		dev_err(&pdev->dev, "failed to attach driver\n");
		goto error_free_pcms;
	}

	/* Did we find a CS42L51 on the I2C bus? */
	if (codec->control_data) {
		/* Initialize codec ops */
		cs42l51_dai.ops.hw_params = cs42l51_hw_params;
		cs42l51_dai.ops.set_sysclk = cs42l51_set_dai_sysclk;
		cs42l51_dai.ops.set_fmt = cs42l51_set_dai_fmt;
	} else
		dev_warn(&pdev->dev, "no I2C device found\n");

	ret = snd_soc_init_card(socdev);
	if (ret < 0) {
		dev_err(&pdev->dev, "failed to register card\n");
		goto error_del_driver;
	}

	return 0;

error_del_driver:
	i2c_del_driver(&cs42l51_i2c_driver);
error_free_pcms:
	snd_soc_free_pcms(socdev);
error_free_codec:
	kfree(socdev->codec);

	socdev->codec = NULL;
	return ret;
}


static int cs42l51_remove(struct platform_device *pdev)
{
	struct snd_soc_device *socdev = platform_get_drvdata(pdev);
	struct snd_soc_codec *codec = socdev->codec;

	if (socdev && socdev->codec) {
		snd_soc_free_pcms(socdev);
		if (codec->control_data) {
			i2c_unregister_device(codec->control_data);
			i2c_del_driver(&cs42l51_i2c_driver);
		}
		kfree(socdev->codec);
		socdev->codec = NULL;
	}

	return 0;
}

struct snd_soc_codec_device soc_codec_device_cs42l51 = {
	.probe =	cs42l51_probe,
	.remove =	cs42l51_remove
};
EXPORT_SYMBOL_GPL(soc_codec_device_cs42l51);

static int __init cs42l51_init(void)
{
	return snd_soc_register_dai(&cs42l51_dai);
}
module_init(cs42l51_init);

static void __exit cs42l51_exit(void)
{
	snd_soc_unregister_dai(&cs42l51_dai);
}
module_exit(cs42l51_exit);

MODULE_AUTHOR("Arnaud Patard <apatard@mandriva.com>");
MODULE_DESCRIPTION("Cirrus Logic CS42L51 ALSA SoC Codec Driver");
MODULE_LICENSE("GPL");

