/*
 * sm501-ac97-dev.c  --  SoC audio for gdium
 *
 * Based on smdk2443_wm9710.c
 *
 * Arnaud Patard <apatard@mandriva.com>
 *
 *  This program is free software; you can redistribute  it and/or modify it
 *  under  the terms of  the GNU General  Public License as published by the
 *  Free Software Foundation;  either version 2 of the  License, or (at your
 *  option) any later version.
 *
 */

#include <linux/module.h>
#include <linux/device.h>
#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/soc.h>
#include <sound/soc-dapm.h>
#include <linux/err.h>
#include <linux/pwm.h>
#include <asm/gpio.h>

#include "../codecs/ac97.h"
#include "sm501-pcm.h"
#include "sm501-ac97.h"

#define SPEAKER_PWM		2
#define SPEAKER_PWM_PERIOD	(1000000000/20000) /* 20 kHz */

#define MUTE_GPIO		192+0 /* pin 32 is mute */
#define AMP_MUTE		1
#define AMP_UNMUTE		0

static struct pwm_device *ap4838_pwm;
static int ap4838_pwm_duty;
static char ap4838_pwm_mute;

static int ap4838_get_vol(struct snd_kcontrol *kcontrol,
		        struct snd_ctl_elem_value *ucontrol)
{
	ucontrol->value.integer.value[0] = ap4838_pwm_duty;
	return 0;
}

static int ap4838_set_vol(struct snd_kcontrol *kcontrol,
		        struct snd_ctl_elem_value *ucontrol)
{
	/* Set duty */
	ap4838_pwm_duty = ucontrol->value.integer.value[0];
	pwm_config(ap4838_pwm, (100-ap4838_pwm_duty) * SPEAKER_PWM_PERIOD / 100, SPEAKER_PWM_PERIOD);

	/* Set Mute */
	if (ap4838_pwm_duty == 0)
		pwm_disable(ap4838_pwm);
	else
		pwm_enable(ap4838_pwm);

	return 1;
}

static int ap4838_set_mute(struct snd_kcontrol *kcontrol,
		        struct snd_ctl_elem_value *ucontrol)
{
        if (ucontrol->value.integer.value[0] == 0) {
		ap4838_pwm_mute = 0;
		gpio_set_value(MUTE_GPIO, AMP_MUTE); /* gpio_direction_input(MUTE_GPIO); */
	}
	else {
		gpio_set_value(MUTE_GPIO, AMP_UNMUTE); /* gpio_direction_output(MUTE_GPIO, 0); */
		ap4838_pwm_mute = 1;
	}

	return 1;
}
static int ap4838_get_mute(struct snd_kcontrol *kcontrol,
		        struct snd_ctl_elem_value *ucontrol)
{
	ucontrol->value.integer.value[0] = ap4838_pwm_mute;
	return 0;
}

static const struct snd_kcontrol_new ap4838_controls[] = {
	SOC_SINGLE_EXT("Amp Playback Volume", 0, 0, 100, 0,
		ap4838_get_vol, ap4838_set_vol),
	SOC_SINGLE_EXT("Amp Playback Switch", 0, 0, 1, 0,
		ap4838_get_mute, ap4838_set_mute),
};

static int ap4838_init(struct snd_soc_codec *codec)
{
	int i, err;

	for (i = 0; i < ARRAY_SIZE(ap4838_controls); i++) {
		err = snd_ctl_add(codec->card,
				snd_soc_cnew(&ap4838_controls[i],
				codec, NULL));
		if (err < 0)
			return err;
	}

	return 0;
}

static int gdium_probe(struct platform_device *pdev)
{
	int ret;

	/* Configure GPIO32 as GPIO */
	ret = gpio_request(MUTE_GPIO, "audio-mute");
	if (ret) {
		dev_err(&pdev->dev, "unable to requestio audio mute gpio\n");
		return ret;
	}
	gpio_set_value(MUTE_GPIO, AMP_MUTE);
	gpio_direction_output(MUTE_GPIO, 1);

	ap4838_pwm = pwm_request(SPEAKER_PWM, "speaker");
	if (ap4838_pwm == NULL) {
		dev_err(&pdev->dev, "unable to request PWM for Gdium speaker\n");
		return -EBUSY;
	}

	pwm_config(ap4838_pwm, 0, SPEAKER_PWM_PERIOD);
	pwm_enable(ap4838_pwm);

	return 0;
}

static int gdium_remove(struct platform_device *pdev)
{
	gpio_free(MUTE_GPIO);

	pwm_config(ap4838_pwm, 0, SPEAKER_PWM_PERIOD);
	pwm_disable(ap4838_pwm);
	pwm_free(ap4838_pwm);

	return 0;
}

#ifdef CONFIG_PM
static int gdium_suspend(struct platform_device *pdev,
				 pm_message_t state)
{
	gpio_set_value(MUTE_GPIO, AMP_MUTE);
	pwm_config(ap4838_pwm, 0, SPEAKER_PWM_PERIOD);
	pwm_disable(ap4838_pwm);
	return 0;
}

static int gdium_resume(struct platform_device *pdev)
{
	pwm_config(ap4838_pwm, (100-ap4838_pwm_duty) * SPEAKER_PWM_PERIOD / 100, SPEAKER_PWM_PERIOD);
	pwm_enable(ap4838_pwm);
	gpio_direction_output(MUTE_GPIO, 1);
	gpio_set_value(MUTE_GPIO, ap4838_pwm_mute ? AMP_UNMUTE : AMP_MUTE);
	return 0;
}
#else
#define gdium_suspend	NULL
#define gdium_resume	NULL
#endif


static struct snd_soc_dai_link gdium_dai[] = {
{
	.name = "AC97",
	.stream_name = "AC97 HiFi",
	.cpu_dai = &sm501_ac97_dai[0],
	.codec_dai = &ac97_dai,
	.init = ap4838_init,
},
};

static struct snd_soc_card gdium_card = {
	.name = "gdium",
	.platform = &sm501_soc_platform,
	.dai_link = gdium_dai,
	.num_links = ARRAY_SIZE(gdium_dai),
	.probe = gdium_probe,
	.remove = gdium_remove,
};

static struct snd_soc_device gdium_snd_devdata = {
	.card = &gdium_card,
	.codec_dev = &soc_codec_dev_ac97,
};

static struct platform_device *gdium_snd_ac97_device;

static void __iomem *remap_mem_resource(struct platform_device *pdev, int resource)
{
	struct resource *res;
	unsigned long res_len;
	void __iomem *ptr = NULL;

	res = platform_get_resource(pdev, IORESOURCE_MEM, resource);
	if (!res) {
		dev_err(&pdev->dev, "No memory resource found\n");
		goto remap_err;
	}

	res_len = (res->end - res->start)+1;
	if (!request_mem_region(res->start, res_len, pdev->name)) {
		dev_err(&pdev->dev, "Can't request iomem resource %lx\n", (unsigned long)res->start);
		goto remap_err;
	}

	ptr = ioremap(res->start, res_len);
	if (!ptr) {
		dev_err(&pdev->dev, "ioremap failed\n");
		goto remap_err_res;
	}

	return ptr;

remap_err_res:
	release_mem_region(res->start, res_len);
remap_err:
	return ptr;
}
static void unmap_mem_resource(struct platform_device *pdev, int resource, void __iomem *regs)
{
	struct resource *res;
	unsigned long res_len;

	res = platform_get_resource(pdev, IORESOURCE_MEM, resource);
	res_len = (res->end - res->start)+1;
	iounmap(regs);
	release_mem_region(res->start, res_len);
}

static __init int snd_gdium_probe(struct platform_device *pdev)
{
	struct sm501_audio *audio_dev;
	int ret;

	audio_dev = kzalloc(sizeof(struct sm501_audio), GFP_KERNEL);
	if (!audio_dev) {
		dev_err(&pdev->dev, "Allocation failure\n");
		return -ENOMEM;
	}
	audio_dev->dev=&pdev->dev;

	gdium_snd_devdata.codec_data = audio_dev;

	platform_set_drvdata(pdev, audio_dev);

	audio_dev->irq = platform_get_irq(pdev, 0);
	if (!audio_dev->irq) {
		dev_err(&pdev->dev, "no irq found\n");
		ret = -ENODEV;
		goto err_alloc;
	}

	audio_dev->ac97_regs = remap_mem_resource(pdev, 0);
	if (!audio_dev->ac97_regs) {
		ret = -EBUSY;
		goto err_alloc;
	}

	audio_dev->u8051_regs = remap_mem_resource(pdev, 1);
	if (!audio_dev->u8051_regs) {
		ret = -EBUSY;
		goto err_map_ac97;
	}

	audio_dev->sram = remap_mem_resource(pdev, 2);
	if (!audio_dev->sram) {
		ret = -EBUSY;
		goto err_map_8051;
	}

	gdium_snd_ac97_device = platform_device_alloc("soc-audio", -1);
	if (!gdium_snd_ac97_device) {
		ret = -ENOMEM;
		goto err_map_sram;
	}

	platform_set_drvdata(gdium_snd_ac97_device,
				&gdium_snd_devdata);
	gdium_snd_devdata.dev = &gdium_snd_ac97_device->dev;

	ret = platform_device_add(gdium_snd_ac97_device);
	if (ret)
		goto err_dev;

	return 0;

err_dev:
	platform_device_put(gdium_snd_ac97_device);
err_map_sram:
	unmap_mem_resource(pdev, 2, audio_dev->sram);
err_map_8051:
	unmap_mem_resource(pdev, 1, audio_dev->u8051_regs);
err_map_ac97:
	unmap_mem_resource(pdev, 0, audio_dev->ac97_regs);
err_alloc:
	kfree(audio_dev);
	return ret;
}

static int snd_gdium_remove(struct platform_device *pdev)
{
	struct sm501_audio *audio_dev = platform_get_drvdata(pdev);

	platform_device_unregister(gdium_snd_ac97_device);
	unmap_mem_resource(pdev, 2, audio_dev->sram);
	unmap_mem_resource(pdev, 1, audio_dev->u8051_regs);
	unmap_mem_resource(pdev, 0, audio_dev->ac97_regs);
	kfree(audio_dev);
	return 0;
}

static struct platform_driver gdium_audio_driver = {
	.driver	= {
		.name	= "sm501-audio",
		.owner	= THIS_MODULE,
	},
	.probe		= snd_gdium_probe,
	.remove		= snd_gdium_remove,
	.suspend	= gdium_suspend,
	.resume		= gdium_resume,
};

static int __init snd_gdium_init(void)
{
	return platform_driver_register(&gdium_audio_driver);
}

static void __exit snd_gdium_exit(void)
{
	platform_driver_unregister(&gdium_audio_driver);
}
module_init(snd_gdium_init);
module_exit(snd_gdium_exit);

/* Module information */
MODULE_AUTHOR("Arnaud Patard <apatard@mandriva.com>");
MODULE_DESCRIPTION("ALSA SoC Gdium");
MODULE_LICENSE("GPL");
MODULE_ALIAS("platform:sm501-audio");
