/*
 * AC97 support for SM50X chipsets
 *
 * Based on Silicon Motion drivers
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/interrupt.h>
#include <linux/wait.h>
#include <linux/delay.h>

#include <sound/core.h>
#include <sound/pcm.h>
#include <sound/ac97_codec.h>
#include <sound/initval.h>
#include <sound/soc.h>

#include <asm/irq.h>
#include <linux/mutex.h>

#include <linux/sm501.h>
#include <linux/sm501-regs.h>

#include "sm501-ac97.h"
#include "sm501-pcm.h"
#include "sm501-u8051.h"
#include "sm501_fw.h"

/*
 * Register address in the firmware
 */
#define SET_TX_SLOT_0                    0
#define SET_TX_SLOT_1                    1
#define SET_TX_SLOT_2                    2
#define SET_PLAYBACK_BUFFER_ADDRESS      3
#define SET_PLAYBACK_BUFFER_SIZE         4
#define SET_PLAYBACK_BUFFER_READY        5
#define GET_BUFFER_STATUS                6
#define START_STOP_AUDIO_PLAYBACK        7
#define GET_RX_SLOT0                     8
#define GET_RX_SLOT1                     9
#define GET_RX_SLOT2                    10
#define SET_CAPTURE_BUFFER_ADDRESS      11
#define SET_CAPTURE_BUFFER_SIZE         12
#define SET_CAPTIRE_BUFFER_EMPTY        13
#define START_STOP_AUDIO_CAPTURE        14
#define SET_GET_AC97_REGISTER           15

#define DRV_NAME			"sm501-ac97"

#define MAKEWORD(l,h)  (((unsigned short) (unsigned char) (l)) | \
			(((unsigned short) (unsigned char) (h)) << 8))
#define MAKEDWORD(l,h) (((unsigned int) (unsigned short) (l)) | \
			(((unsigned int) (unsigned short) (h)) << 16))
#define SRAM(x)      (offsetof(FIRMWARE_SRAM, x))

static int debug        = 0;
#define dprintk(msg...) if (debug) { printk(KERN_DEBUG DRV_NAME ": " msg); }
static struct sm501_audio *audio_dev;

static void sm501_write_codec(struct sm501_audio *s, u8 addr, u16 data)
{
	unsigned int read_back;

	dprintk("%s: addr=%02x data=%04x\n", __func__, addr, data);

	/* Write Codec register. */
	if (!sm501_write_command(s, SET_GET_AC97_REGISTER, (unsigned int) addr << 12,
	     (unsigned int) data << 4) && (addr != AC97_RESET)) {
		read_back = MAKEDWORD(MAKEWORD(readb(audio_dev->sram+SRAM(data_byte[4])),
					readb(audio_dev->sram+SRAM(data_byte[5]))),
					MAKEWORD(readb(audio_dev->sram+SRAM(data_byte[6])),
					readb(audio_dev->sram + SRAM(data_byte[7])))) >> 4;

		/*
		 * Not dev_err here because the firmware returns an error
		 * for R/O or not defined registers
		 */
		dev_dbg(s->dev, "%s failed: addr=%02x data=%04x read_back=%04x\n",
			__func__, addr, data, read_back);
	}
}

static u16 sm501_read_codec(struct sm501_audio *s, u8 addr)
{
	unsigned int data;
	int i;

	/* Read Codec register. */
	if (sm501_write_command(s, SET_GET_AC97_REGISTER,
	    (unsigned int) (addr << 12) | (1 << 19), 0)) {
		data = MAKEDWORD(MAKEWORD(readb(audio_dev->sram+SRAM(data_byte[4])),
				readb(audio_dev->sram+SRAM(data_byte[5]))),
				MAKEWORD(readb(audio_dev->sram+SRAM(data_byte[6])),
				readb(audio_dev->sram+SRAM(data_byte[7])))) >> 4;
	} else {
		/* Read register from SRAM in case link failes. */
		for (i = SM501_AC97_CODEC_TIMEOUT; i > 0; i--) {
			data = readw(audio_dev->sram+SRAM(ac97_regs[addr >> 1]));
			if (data != 0xFFFF)
				break;
			msleep(1);
		}

		dev_err(s->dev, "%s failed: addr=%02x sram=%04x\n", __func__, addr, data);
	}

	dprintk("%s: data=%04x\n", __func__, data);
	return((unsigned short) data);
}

/*
 * AC97 functions
 */
static unsigned short snd_sm501_ac97_read(struct snd_ac97 *ac97, unsigned short reg)
{
	unsigned int data;
	unsigned long value;
	unsigned long flags;
	unsigned long i;

	spin_lock_irqsave(&audio_dev->lock, flags);

	if (audio_dev->firmware_loaded) {
		data = sm501_read_codec(audio_dev, reg);
	} else {

		writel(0, audio_dev->ac97_regs + AC97_TX_SLOT2);

		value = (unsigned long) reg << 12;
		value |= (1<<19);
		writel(value, audio_dev->ac97_regs + AC97_TX_SLOT1);

		/* Enable Slot 0, 1, 2 for read */
		value = readl(audio_dev->ac97_regs+AC97_TX_SLOT0);
		value |= (1<<15);
		value |= (1<<14);
		value |= (1<<13);
		writel(value, audio_dev->ac97_regs + AC97_TX_SLOT0);

		udelay(150);

		for (i = SM501_AC97_CODEC_TIMEOUT; i> 0; i--)
		{
			if (readl(audio_dev->ac97_regs + AC97_RX_SLOT0) & (1<<15))
			{
				value = readl(audio_dev->ac97_regs + AC97_RX_SLOT0);
				if ((value & (1<<14)) && (value & (1<<13)))
					break;
			}
			udelay(1);
		}
		if (i == 0)
			dev_err(audio_dev->dev, "%s: AC97 timeout\n", __func__);


		data = readl(audio_dev->ac97_regs + AC97_RX_SLOT2);
		data = (data>>4) & 0xffff;

		/* Reset Slot 1,2 */
		value = readl(audio_dev->ac97_regs + AC97_TX_SLOT0);
		value &= ~(1<<14);
		value &= ~(1<<13);
		writel(value, audio_dev->ac97_regs + AC97_TX_SLOT0);
	}

	spin_unlock_irqrestore(&audio_dev->lock, flags);

	return((unsigned short) data);
}

static void snd_sm501_ac97_write(struct snd_ac97 *ac97, unsigned short reg,
				 unsigned short val)
{
	unsigned long flags;
	unsigned long value;
	unsigned long i;

	spin_lock_irqsave(&audio_dev->lock, flags);

	if (audio_dev->firmware_loaded) {
		sm501_write_codec(audio_dev, reg, val);
	} else {
		value = (unsigned long) reg << 12;
		writel(value, audio_dev->ac97_regs + AC97_TX_SLOT1);

		value = ((unsigned int) val << 4);
		writel(value, audio_dev->ac97_regs + AC97_TX_SLOT2);

		/* Enable Slot 0, 1, 2 for write */
		value = readl(audio_dev->ac97_regs + AC97_TX_SLOT0);
		value |= (1<<15);
		value |= (1<<14);
		value |= (1<<13);
		writel(value, audio_dev->ac97_regs + AC97_TX_SLOT0);

		for (i = SM501_AC97_CODEC_TIMEOUT; i> 0; i--)
		{
			if (readl(audio_dev->ac97_regs + AC97_RX_SLOT0) & (1<<15))
			{
				value = readl(audio_dev->ac97_regs + AC97_RX_SLOT0);
				if (!(value & (1<<14)) && !(value & (1<<13)))
					break;
			}
			udelay(1);
		}
		if (i == 0)
			dev_err(audio_dev->dev, "%s: AC97 timeout\n", __func__);

		/* Reset Slot 1,2 */
		value = readl(audio_dev->ac97_regs + AC97_TX_SLOT0);
		value &= ~(1<<14);
		value &= ~(1<<13);
		writel(value, audio_dev->ac97_regs + AC97_TX_SLOT0);
	}

	spin_unlock_irqrestore(&audio_dev->lock, flags);
}

static void snd_sm501_ac97_warm_reset(struct snd_ac97 *ac97)
{
#warning FILL ME
}

static void snd_sm501_ac97_cold_reset(struct snd_ac97 *ac97)
{
#warning FILL ME
}

static void snd_sm501_ac97_init(struct snd_ac97 *ac97)
{
#ifdef CONFIG_LOONGSON2F_GDIUM
	/* TODO: find better ids */
	ac97->subsystem_vendor = 0xdead;
	ac97->subsystem_device = 0xbeef;
#endif
}

struct snd_ac97_bus_ops soc_ac97_ops = {
	.init		= snd_sm501_ac97_init,
	.read		= snd_sm501_ac97_read,
	.write		= snd_sm501_ac97_write,
	.warm_reset	= snd_sm501_ac97_warm_reset,
	.reset		= snd_sm501_ac97_cold_reset,
};
EXPORT_SYMBOL_GPL(soc_ac97_ops);

static struct sm501_pcm_dma_params sm501_ac97_pcm_stereo_out;
static struct sm501_pcm_dma_params sm501_ac97_pcm_stereo_in;

/* hw_params callback */
static int snd_sm501_ac97_hw_params(struct snd_pcm_substream *substream,
				struct snd_pcm_hw_params *params, struct snd_soc_dai *cpu_dai)
{
	if (substream->stream == SNDRV_PCM_STREAM_PLAYBACK)
		cpu_dai->dma_data = &sm501_ac97_pcm_stereo_out;
	else
		cpu_dai->dma_data = &sm501_ac97_pcm_stereo_in;

	return 0;
}
static int snd_sm501_ac97_prepare(struct snd_pcm_substream *substream, struct snd_soc_dai *dai)
{
	struct snd_pcm_runtime *runtime = substream->runtime;
	int reg;
	u16 vra;

	vra = sm501_read_codec(audio_dev, AC97_EXTENDED_STATUS);
	sm501_write_codec(audio_dev, AC97_EXTENDED_STATUS, vra | AC97_EA_VRA);

	if (substream->stream == SNDRV_PCM_STREAM_PLAYBACK)
		reg = AC97_PCM_FRONT_DAC_RATE;
	 else
		reg = AC97_PCM_LR_ADC_RATE;

	 sm501_write_codec(audio_dev, reg, runtime->rate);
	 return 0;
}

static int snd_sm501_ac97_configure(void)
{
	unsigned long value;
	int ret;

	sm501_configure_gpio(audio_dev->dev->parent, 28, 1);
	sm501_configure_gpio(audio_dev->dev->parent, 27, 1);
	sm501_configure_gpio(audio_dev->dev->parent, 26, 1);
	sm501_configure_gpio(audio_dev->dev->parent, 25, 1);
	sm501_configure_gpio(audio_dev->dev->parent, 24, 1);

	/* Initialize chip. */
	sm501_unit_power(audio_dev->dev->parent, SM501_GATE_AC97, 1);
	sm501_unit_power(audio_dev->dev->parent, SM501_GATE_UCONTROLLER, 1);

	/* Set arbitartion control */
	sm501_modify_reg(audio_dev->dev->parent, SM501_ARBTRTN_CONTROL,(1<<8),(7<<8));

	/* Enable interrupts */
	sm501_modify_reg(audio_dev->dev->parent, SM501_IRQ_MASK, (1<<10)|(1<<20), 0);


	value = readl(audio_dev->u8051_regs + SM501_UCONTROLLER_MODE_SELECT);
	value &= ~(1<<5);
	value &= ~(1<<3);
	value &= ~(1<<2);
	value &= ~(0x3);
	writel(value, audio_dev->u8051_regs + SM501_UCONTROLLER_MODE_SELECT);

	value = readl(audio_dev->ac97_regs + AC97_CONTROL_STATUS);
	value |= (1<<0);
	writel(value, audio_dev->ac97_regs + AC97_CONTROL_STATUS);

	/*
	 * We're failing here but one should be able to use the codec
	 * without firmware according to the manual
	 */
	audio_dev->firmware_loaded = 0;
	ret = sm501_load_firmware(audio_dev, u8051_firmware);
	if (ret < 0) {
		dev_err(audio_dev->dev, "Error while loading firmware (%d)\n", ret);
	}
	else
		audio_dev->firmware_loaded = 1;
	return ret;
}

static int snd_sm501_ac97_probe(struct platform_device *pdev, struct snd_soc_dai *dai)
{
	struct snd_soc_device *socdev = platform_get_drvdata(pdev);
	struct device   *dev = &pdev->dev;
	int ret = 0;

	audio_dev = socdev->codec_data;
	sm501_ac97_pcm_stereo_out.s = audio_dev;
	sm501_ac97_pcm_stereo_in.s = audio_dev;
	spin_lock_init(&audio_dev->lock);

	ret = snd_sm501_ac97_configure();
	if (ret) {
		dev_err(dev, "configure failed\n");
		goto err;
	}

	return 0;

err:
	return ret;
}

static void snd_sm501_ac97_remove(struct platform_device *pdev, struct snd_soc_dai *dai)
{
	unsigned int value;

	/* Disable 8051 interrupt mask. */
	sm501_modify_reg(audio_dev->dev->parent, SM501_IRQ_MASK, 0, (1<<10));


	/* Put 8051 in reset state. */
	value = readl(audio_dev->u8051_regs + SM501_UCONTROLLER_RESET);
	value &= ~(1<<0);
	writel(value, audio_dev->u8051_regs + SM501_UCONTROLLER_RESET);

	/* Disable SRAM. */
	value = readl(audio_dev->u8051_regs + SM501_UCONTROLLER_MODE_SELECT);
	value |= (1<<5);
	writel(value, audio_dev->u8051_regs + SM501_UCONTROLLER_MODE_SELECT);

	/* Reset GPIO lines. */
	sm501_configure_gpio(audio_dev->dev->parent, 28, 0);
	sm501_configure_gpio(audio_dev->dev->parent, 27, 0);
	sm501_configure_gpio(audio_dev->dev->parent, 26, 0);
	sm501_configure_gpio(audio_dev->dev->parent, 25, 0);
	sm501_configure_gpio(audio_dev->dev->parent, 24, 0);

	/* Disable 8051 and AC97. */
	sm501_unit_power(audio_dev->dev->parent, SM501_GATE_AC97, 0);
	sm501_unit_power(audio_dev->dev->parent, SM501_GATE_UCONTROLLER, 0);
}

#ifdef CONFIG_PM
static int snd_sm501_ac97_suspend(struct snd_soc_dai *cpu_dai)
{
	unsigned int value;

	audio_dev->firmware_loaded = 0;

	/* Disable 8051 interrupt mask. */
	sm501_modify_reg(audio_dev->dev->parent, SM501_IRQ_MASK, 0, (1<<10));


	/* Put 8051 in reset state. */
	value = readl(audio_dev->u8051_regs + SM501_UCONTROLLER_RESET);
	value &= ~(1<<0);
	writel(value, audio_dev->u8051_regs + SM501_UCONTROLLER_RESET);

	/* Disable SRAM. */
	value = readl(audio_dev->u8051_regs + SM501_UCONTROLLER_MODE_SELECT);
	value |= (1<<5);
	writel(value, audio_dev->u8051_regs + SM501_UCONTROLLER_MODE_SELECT);

	/* Disable 8051 and AC97. */
	sm501_unit_power(audio_dev->dev->parent, SM501_GATE_AC97, 0);
	sm501_unit_power(audio_dev->dev->parent, SM501_GATE_UCONTROLLER, 0);

	return 0;
}

static int snd_sm501_ac97_resume(struct snd_soc_dai *cpu_dai)
{
	return snd_sm501_ac97_configure();
}
#else
#define snd_sm501_ac97_suspend	NULL
#define snd_sm501_ac97_resume	NULL
#endif

struct snd_soc_dai sm501_ac97_dai[] = {{
	.name = DRV_NAME,
	.id = 0,
	.ac97_control = 1,
	.probe = snd_sm501_ac97_probe,
	.remove = snd_sm501_ac97_remove,
	.suspend = snd_sm501_ac97_suspend,
	.resume = snd_sm501_ac97_resume,
	.playback = {
		.stream_name = "AC97 Playback",
		.channels_min = 2,
		.channels_max = 2,
		.rates = SNDRV_PCM_RATE_48000, /* SNDRV_PCM_RATE_8000_48000,*/
		/* change to be with be systems */
		.formats = SNDRV_PCM_FMTBIT_S16_LE,},
	.capture = {
		.stream_name = "AC97 Capture",
		.channels_min = 2,
		.channels_max = 2,
		.rates = SNDRV_PCM_RATE_48000,
		.formats = SNDRV_PCM_FMTBIT_S16_LE,},
	.ops = {
		.hw_params = snd_sm501_ac97_hw_params,
		.prepare   = snd_sm501_ac97_prepare,
	},
},
#if 0
#warning Put mic here
{}
#endif
};
EXPORT_SYMBOL_GPL(sm501_ac97_dai);

static int __devinit sm501_ac97_init(void)
{
	return snd_soc_register_dais(sm501_ac97_dai, ARRAY_SIZE(sm501_ac97_dai));
}
module_init(sm501_ac97_init);

static void __exit sm501_ac97_exit(void)
{
	snd_soc_unregister_dais(sm501_ac97_dai, ARRAY_SIZE(sm501_ac97_dai));
}
module_exit(sm501_ac97_exit);

MODULE_AUTHOR("Arnaud Patard <apatard@mandriva.com>");
MODULE_DESCRIPTION("ASoC AC97 driver for SM501 chip");
MODULE_LICENSE("GPL");
